#include "aggr.h"

#include "disco.h"
#include "flow.h"
#include <string.h>
#include <stdio.h>

aggregator_t *init_aggregator(int num_fields, char **field_names) {
  aggregator_t *aggr = (aggregator_t *)malloc(sizeof(aggregator_t));
  aggr->num_fields = num_fields;
  aggr->key_length = 0;

  const flow_field **fields = (const flow_field **)malloc(num_fields * sizeof(flow_field *));
  for(int i = 0; i < num_fields; i++) {
    for(const flow_field *f = flow_fields; f->name; f++) {
      if(strcmp(field_names[i], f->name) == 0) {
        fields[i] = f;
        aggr->key_length += f->length;
        break;
      }
    }
    if(!fields[i]) {
      fprintf(stderr, "Unrecognized field name: %s\n", field_names[i]);
      exit(1);
    }
  }
  aggr->fields = fields;

  return aggr;
}

aggregate_t *aggregate_alloc(aggregator_t *aggregator) {
  if ((aggregator->count % AGGREGATE_SLAB_SIZE) == 0) {
    /* We need a new slab. */
    struct aggregate_slab_s *s = malloc(sizeof(struct aggregate_slab_s));
    memset(s, 0, sizeof(*s));

    s->next = aggregator->slab;
    aggregator->slab = s;
  }

  aggregate_t *a = &aggregator->slab->aggregates[aggregator->count % AGGREGATE_SLAB_SIZE];
  aggregator->count += 1;

  return a;
}

uint8_t *aggregate_key(aggregator_t *aggr, flow_t *flow) {
  uint8_t *key = (uint8_t *) malloc(aggr->key_length * sizeof(uint8_t));
  uint8_t *os = key;
  for(int i = 0; i < aggr->num_fields; i++) {
    memcpy(os, (void *)flow + aggr->fields[i]->offset, aggr->fields[i]->length);
    os += aggr->fields[i]->length;
  }
  return key;
}
