#include "aggr.h"

#include "flow.h"
#include <string.h>
#include <stdio.h>

aggregator_t *init_aggregator(int num_fields, char **field_names) {
  aggregator_t *aggr = (aggregator_t *)malloc(sizeof(aggregator_t));
  aggr->num_fields = num_fields;
  aggr->key_length = 0;
  aggr->index = (PWord_t) NULL;
  aggr->count = 0;

  const flow_field **fields = (const flow_field **)malloc(num_fields * sizeof(flow_field *));
  for(int i = 0; i < num_fields; i++) {
    for(const flow_field *f = flow_fields; f->name; f++) {
      if(strcmp(field_names[i], f->name) == 0) {
        fields[i] = f;
        aggr->key_length += f->length;
        break;
      }
    }
    if(!fields[i]) {
      fprintf(stderr, "Unrecognized field name: %s\n", field_names[i]);
      exit(1);
    }
  }
  aggr->fields = fields;

  /* Set up buffer for assembling keys from flow fields */
  aggr->key_buffer = (uint8_t *)malloc(aggr->key_length);
  return aggr;
}

void *aggregate_alloc(aggregator_t *aggregator) {
  if ((aggregator->count % AGGREGATE_SLAB_SIZE) == 0) {
    /* We need a new slab. */
    struct aggregate_slab_s *s = malloc(sizeof(struct aggregate_slab_s));
    memset(s, 0, sizeof(*s));

    void *a = malloc((sizeof(aggregate_t) + aggregator->key_length) * AGGREGATE_SLAB_SIZE);

    s->next = aggregator->slab;
    s->aggregates = a;
    aggregator->slab = s;
  }

  size_t aggregate_num = aggregator->count % AGGREGATE_SLAB_SIZE;
  size_t aggregate_size = sizeof(aggregate_t) + aggregator->key_length;
  void *a = aggregator->slab->aggregates + (aggregate_size * aggregate_num);

  aggregator->count += 1;

  return (aggregate_t *)a;
}

void aggregate_flow(aggregator_t *aggregator, flow_t *flow) {
  uint8_t *key, *os;
  PWord_t aggr_lookup;

  /* TODO Optimize malloc away */
  key = (uint8_t *) malloc(aggregator->key_length * sizeof(uint8_t));
  os = key;
  for(int i = 0; i < aggregator->num_fields; i++) {
    memcpy(os, (void *)flow + aggregator->fields[i]->offset, aggregator->fields[i]->length);
    os += aggregator->fields[i]->length;
  }

  JHSI(aggr_lookup, aggregator->index, key, aggregator->key_length);
  if(*aggr_lookup == 0) {
    void *a = aggregate_alloc(aggregator);

    uint8_t *aggr_key = KEY(a);
    memcpy(aggr_key, key, aggregator->key_length);

    aggregate_t *aggr = AGGREGATE(a);
    aggr->flows = 1;
    aggr->packets = flow->packets;
    aggr->bytes = flow->bytes;

    *aggr_lookup = (Word_t)a;
  } else {
    aggregate_t *aggr = AGGREGATE(*aggr_lookup);
    aggr->flows++;
    aggr->packets += flow->packets;
    aggr->bytes += flow->bytes;
  }

  free(key);
}

void aggregate_merge(aggregator_t *aggregator, void *aggr) {
  PWord_t aggr_lookup;

  JHSI(aggr_lookup, aggregator->index, KEY(aggr), aggregator->key_length);
  if(*aggr_lookup == 0) {
    void *a = aggregate_alloc(aggregator);
    memcpy(a, aggr, sizeof(aggregate_t) + aggregator->key_length);
    *aggr_lookup = (Word_t) a;
  } else {
    aggregate_t *old_aggr = AGGREGATE(*aggr_lookup);
    aggregate_t *new_aggr = AGGREGATE(aggr);
    old_aggr->flows += new_aggr->flows;
    old_aggr->packets += new_aggr->packets;
    old_aggr->bytes += new_aggr->bytes;
  }

}
