#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "aggr.h"
#include "flow.h"
#include "stream.h"
#include "sink.h"

#define READ_ENTRIES 78

void filter_stdin(aggregator_t *a) {
  flow_t *buf = (flow_t *) malloc(sizeof(flow_t) * READ_ENTRIES);
  size_t num;

  for(;;) {
    num = fread(buf, sizeof(flow_t), READ_ENTRIES, stdin);

    for(int i = 0; i < num; i++) {
      flow_t *f = &buf[i];
      aggregate_flow(a, f);
    }

    if(num < READ_ENTRIES) {
      break;
    }
  }
}

void filter_stdout(aggregator_t *aggregator) {
  size_t count = aggregator->count % AGGREGATE_SLAB_SIZE;
  aggregate_slab_t *slab;

  for (slab = aggregator->slab; slab != NULL; slab = slab->next, count = AGGREGATE_SLAB_SIZE) {
    size_t written = 0;
    while(written < count) {
      written += fwrite(slab->aggregates + written, sizeof(aggregate_t) + aggregator->key_length, count, stdout);
    }
  }
}

int main(int argc, char **argv) {
  int stdout = 0, nodes = NUM_NODES;

  int c;
  int opterr = 0;
  while((c = getopt(argc, argv, "cn:")) != -1) {
    switch(c)
      {
      case 'c':
        stdout = 1;
        break;
      case 'n':
        nodes = atoi(optarg);
        break;
      case '?':
        if(optopt == 'n')
          fprintf(stderr, "Missing argument to -%c\n", optopt);
        else
          fprintf(stderr, "Unknown option -%c\n", optopt);
        opterr++;
      default:
        opterr++;
      }
  }

  if(opterr || optind >= argc) {
    fprintf(stderr, "Usage: %s [-c] [-n nodes] field [field ...]\n", argv[0]);
    exit(1);
  }

  if(!stdout && (nodes < 1 || nodes > NUM_NODES)) {
    fprintf(stderr, "Number of nodes (%i) either too small or exceeds maximum of %i\n", nodes, NUM_NODES);
    exit(1);
  }


  aggregator_t *aggregator = init_aggregator(argc - optind, argv + optind);

  /* Read and aggregate data from stdin */
  filter_stdin(aggregator);

  fprintf(stderr, "Total aggregates: %zi\n", aggregator->count);

  if(stdout == 1) {
    filter_stdout(aggregator);
    exit(0);
  }

  /* Distribute aggregates to aggregator nodes */
  sink_init(aggregator, nodes);

  fd_set fds;
  for(;;) {
    FD_ZERO(&fds);
    int max = sink_add_fds(&fds);

    if(max == -1) {
      /* All connections closed */
      break;
    }

    struct timeval timeout = {1, 0};
    int count = select(max + 1, NULL, &fds, NULL, &timeout);

    if(count == -1 && errno == EINTR) {
      continue;
    }

    sink_process(&fds, aggregator);
  }

  return 0;
}
