#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <fcntl.h>

#include "aggr.h"
#include "source.h"

#define NUM_NODES 15

static int listen_fd;

void server_process(fd_set *fds) {
  if(!FD_ISSET(listen_fd, fds)) {
    return;
  }

  struct sockaddr_in ina;
  socklen_t len = sizeof(ina);
  int fd = accept(listen_fd, (struct sockaddr *) &ina, &len);

  int flags = fcntl(fd, F_GETFL);
  int ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
  if(ret == -1) {
    fprintf(stderr, "Couldn't set non-blocking: %s\n", strerror(errno));
    exit(1);
  }

  source_add(fd);
}

void server_init(int port, aggregator_t *a) {
  listen_fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(listen_fd == -1) {
    fprintf(stderr, "Error creating listen socket: %s\n", strerror(errno));
    exit(1);
  }

  int one = 1;
  if(setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
    fprintf(stderr, "Error enabling SO_REUSEADDR on server socket: %s\n", strerror(errno));
    exit(1);
  }

  struct sockaddr_in listen_addr;
  listen_addr.sin_family = AF_INET;
  listen_addr.sin_addr.s_addr = INADDR_ANY;
  listen_addr.sin_port = htons(port);

  if(bind(listen_fd, (struct sockaddr *)&listen_addr, sizeof(listen_addr)) == -1) {
    fprintf(stderr, "Error binding: %s\n", strerror(errno));
    exit(1);
  }

  if(listen(listen_fd, NUM_NODES) == -1) {
    fprintf(stderr, "Error listening: %s\n", strerror(errno));
    exit(1);
  }

  source_init();

  fd_set fds;

  for(;;) {
    FD_ZERO(&fds);
    FD_SET(listen_fd, &fds);

    struct timeval timeout = {1, 0};
    int max = listen_fd;

    int source_max = source_add_fds(&fds);
    if(source_max > max)
      max = source_max;

    int count = select(max + 1, &fds, NULL, NULL, &timeout);

    if(count == -1) {
      if(errno == EINTR) {
        continue;
      }

      fprintf(stderr, "Select failed: %s\n", strerror(errno));
      exit(1);
    }

    server_process(&fds);
    source_process(&fds, a);
  }
}
