#include "source.h"
#include "flow.h"
#include "aggr.h"
#include "server.h"

#include <unistd.h>
#include <string.h>

static source_t *sources = NULL;

void source_init(void) {
  sources = malloc(sizeof(source_t) * NUM_NODES);
  for(int i = 0; i < NUM_NODES; i++) {
    sources[i].fd = -1;
  }
}

void source_add(int fd) {
  for(int i = 0; i < NUM_NODES; i++) {
    if(sources[i].fd == -1) {
      sources[i].fd = fd;
      sources[i].buffer = (char *)malloc(BUFFER_SIZE);
      sources[i].pos = 0;
      break;
    }
  }
}

void source_close(source_t *source) {
  close(source->fd);
  source->fd = -1;
  source->pos = 0;
}

int source_add_fds(fd_set *fds) {
  int max = -1;
  for(int i = 0; i < NUM_NODES; i++) {
    if(sources[i].fd != -1) {
      FD_SET(sources[i].fd, fds);
      if(sources[i].fd > max) {
        max = sources[i].fd;
      }
    }
  }
  return max;
}

void source_process(fd_set *fds, aggregator_t *a) {
  for(int i = 0; i < NUM_NODES; i++) {
    source_t *s = &sources[i];

    if(s->fd == -1)
      continue;

    if(FD_ISSET(s->fd, fds)) {

      /* Read data from source */
      ssize_t len = read(s->fd, s->buffer + s->pos, BUFFER_SIZE - s->pos);

      /* Close connection if we read nothing */
      if(len == 0) {
        source_close(s);
        continue;
      }

      //fprintf(stderr, "[%i] read %i bytes\n", s->fd, len);
      char *end = s->buffer + s->pos + len;

      /* Process all complete records */
      char *pos = s->buffer;
      //fprintf(stderr, "[%i] pos %i, len %i\n", s->fd, s->pos, len);
      while(pos + sizeof(aggregate_t) + a->key_length <= end) {
        //fprintf(stderr, "[%i] processing aggregate\n", s->fd);
        aggregate_merge(a, (void *)pos);
        //fprintf(stderr, "[%i] now at %zi aggregates\n", s->fd, a->count);
        pos += sizeof(aggregate_t) + a->key_length;
      }

      /* Shift any partial, unread record to the beginning of the buffer */
      /* TODO Check whether areas overlap and memcpy() if they don't */
      memmove(s->buffer, pos, end - pos);
      s->pos = end - pos;
    }
  }
}
